#include "stdafx.h"
#include "twainman.h"
#include <math.h>
#include<float.h>


#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

#define lengthof(X) (sizeof(X)/sizeof(X[0]))

#ifdef _DEBUG
void TRACE(const TCHAR *pFormat, ...)
{
    TCHAR	szBuff[1024];
    va_list	argptr;

    va_start(argptr, pFormat);
    _vstprintf_s(szBuff, lengthof(szBuff), pFormat, argptr);
    va_end(argptr);

    OutputDebugString(szBuff);

    return;
}
#else
#define TRACE
#endif




CTWAINManager::CTWAINManager()
{
	m_hParentWnd = NULL;
	m_pfnCallBack = NULL;
	m_bEnableDS = FALSE;
	m_bOpenDSM = FALSE;
	m_bOpenDS = FALSE;
	m_pDSM_Entry = NULL;
	m_hDSMLib = NULL;
	memset(&m_DSID, 0, sizeof(TW_IDENTITY));
	memset (&m_GlobalStatus, 0, sizeof(TW_STATUS));
}

CTWAINManager::~CTWAINManager()
{
	if(m_hDSMLib){
		FreeLibrary(m_hDSMLib);
	}
}

BOOL CTWAINManager::ExitTWAIN()
{
	if(m_hDSMLib){
		BOOL bRet = FreeLibrary(m_hDSMLib);
		m_hDSMLib = NULL;
		m_pDSM_Entry = NULL;

	}
	return FALSE;
}

BOOL CTWAINManager::InitTWAIN()
{
	if(m_pDSM_Entry == NULL && m_hDSMLib == NULL){
		if((m_hDSMLib = LoadLibrary(_T("TWAIN_32.DLL"))) != NULL){
			if((m_pDSM_Entry = (DSMENTRYPROC)GetProcAddress(m_hDSMLib, (LPCSTR)MAKEINTRESOURCE(1))) == NULL){
				FreeLibrary(m_hDSMLib);
				return FALSE;
			}else{
				m_AppID.Id = 0; 				// init to 0, but Source Manager will assign real value
				m_AppID.Version.MajorNum = 1;
				m_AppID.Version.MinorNum = 0;
				m_AppID.Version.Language = TWLG_JAPANESE;
				m_AppID.Version.Country  = TWCY_JAPAN;
				strcpy (m_AppID.Version.Info,  "PaintTail");
				strcpy (m_AppID.ProductName,   "PaintTail");
				m_AppID.ProtocolMajor = TWON_PROTOCOLMAJOR;
				m_AppID.ProtocolMinor = TWON_PROTOCOLMINOR;
				m_AppID.SupportedGroups =  DG_IMAGE | DG_CONTROL;
				strcpy (m_AppID.Manufacturer,  "ran-yu soft");
				strcpy (m_AppID.ProductFamily, "PaintTail");
				return TRUE;
			}
		}
	}else{
		TRACE(_T("Already open\n"));
	}
	return FALSE;
}

BOOL CTWAINManager::IsUsableTWAIN()
{ 
	return (m_pDSM_Entry != NULL); 
}



TW_UINT16 CTWAINManager::CallDSMEntry(TW_IDENTITY* pApp, TW_IDENTITY* pSrc,
										TW_UINT32 DG, TW_UINT16 DAT, TW_UINT16 MSG, TW_MEMREF pData)
{
	if(!IsUsableTWAIN())	{ return TWCC_UNUSABLETWAIN; }

	TW_UINT16 twRC = (*m_pDSM_Entry)(pApp, pSrc, DG, DAT, MSG, pData);
	if((twRC != TWRC_SUCCESS)&&(DAT!=DAT_EVENT))
	{
		if((*m_pDSM_Entry)(pApp, pSrc, DG_CONTROL, DAT_STATUS, MSG_GET, 
					(TW_MEMREF)&m_GlobalStatus) == TWRC_SUCCESS)
		TRACE(_T("CallDSMEntry function: call failed with RC = %d, CC = %d.\n"), 
					twRC, m_GlobalStatus.ConditionCode);
	}else{
		m_GlobalStatus.ConditionCode = TWRC_SUCCESS;
	}
	return twRC;
}


TW_UINT16 CTWAINManager::OpenDSM(HWND hWnd)
{
	if(!IsUsableTWAIN())	return TWCC_UNUSABLETWAIN;
	if(m_bOpenDSM)	return TWCC_DSM_IS_ALREADYOPEN;
	TW_UINT16 twRC = CallDSMEntry(&m_AppID, NULL, DG_CONTROL, DAT_PARENT, MSG_OPENDSM, (TW_MEMREF)&hWnd);
	m_bOpenDSM = (twRC == TWRC_SUCCESS);
	return twRC;
}


TW_UINT16  CTWAINManager::CloseDSM(HWND hWnd)
{
	if(!IsUsableTWAIN())	return TWCC_UNUSABLETWAIN;
	if(!m_bOpenDSM)			return TWCC_DSM_IS_NOTOPEN;
	m_bOpenDSM = FALSE;
	return CallDSMEntry(&m_AppID, NULL, DG_CONTROL, DAT_PARENT, MSG_CLOSEDSM, (TW_MEMREF)&hWnd);
}



TW_UINT16 CTWAINManager::SelectDS ()
{
	if(!IsUsableTWAIN())	return TWCC_UNUSABLETWAIN;
	TW_IDENTITY newDSID;
	memset(&newDSID, 0, sizeof(TW_IDENTITY));
	TW_UINT16 twRC;
	//ꉞftHgƂĂ
	twRC = CallDSMEntry(&m_AppID, NULL, DG_CONTROL, DAT_IDENTITY, MSG_GETDEFAULT, (TW_MEMREF)&newDSID);
	//I
	twRC = CallDSMEntry(&m_AppID, NULL, DG_CONTROL, DAT_IDENTITY, MSG_USERSELECT, (TW_MEMREF)&newDSID);
	if(twRC == TWRC_SUCCESS){
		m_DSID = newDSID; 
	}
	return twRC;
}

//
TW_UINT16 CTWAINManager::OpenDS()
{
	if(!IsUsableTWAIN())	return TWCC_UNUSABLETWAIN;
	if(m_DSID.Id == 0){//not yet select source
		strcpy(m_DSID.ProductName, "\0");
	}
	if (m_bOpenDS)		return TWCC_DS_IS_ALREADYOPEN;
	TW_UINT16 twRC = CallDSMEntry(&m_AppID, NULL, DG_CONTROL, DAT_IDENTITY, MSG_OPENDS, (TW_MEMREF)&m_DSID);
	if(twRC == TWRC_SUCCESS){
		m_bOpenDS = TRUE;
	}
	return twRC;
}

TW_UINT16 CTWAINManager::CloseDS()
{
	if(!IsUsableTWAIN())	return TWCC_UNUSABLETWAIN;
	m_bOpenDS = FALSE;
	return CallDSMEntry(&m_AppID, NULL, DG_CONTROL, DAT_IDENTITY, MSG_CLOSEDS, (TW_MEMREF)&m_DSID);
}

TW_UINT16 CTWAINManager::EnableDS(HWND hWnd)
{
	if(!IsUsableTWAIN())	return TWCC_UNUSABLETWAIN;
	if(m_bEnableDS)	return TWCC_DS_IS_ALREADYENABLE;
	TW_USERINTERFACE ui;
	ui.ShowUI = TRUE;
	ui.ModalUI = TRUE;
	ui.hParent = hWnd;
	TW_UINT16 twRC = CallDSMEntry(&m_AppID, &m_DSID, DG_CONTROL, DAT_USERINTERFACE, MSG_ENABLEDS, (TW_MEMREF)&ui);
	if(twRC == TWRC_SUCCESS){
		::EnableWindow(hWnd, FALSE);
		m_bEnableDS = TRUE;
	}
	return twRC;
}

TW_UINT16 CTWAINManager::DisableDS(HWND hWnd)
{
	if(!IsUsableTWAIN())	return TWCC_UNUSABLETWAIN;
	if(!m_bEnableDS)	return TWCC_DS_IS_NOTENABLE;
	TW_USERINTERFACE ui;
	TW_UINT16 twRC = CallDSMEntry(&m_AppID, &m_DSID, DG_CONTROL, DAT_USERINTERFACE, MSG_DISABLEDS, (TW_MEMREF)&ui);
	m_bEnableDS = FALSE;
	::EnableWindow(hWnd, TRUE);
	return twRC;
}

TW_UINT16 CTWAINManager::SelectScaner(HWND hWnd)
{
	TW_UINT16 twRC = TWCC_UNUSABLETWAIN;
	if(IsUsableTWAIN()){
		if((twRC =OpenDSM(hWnd)) == TWRC_SUCCESS){
			twRC =SelectDS();
			CloseDSM(hWnd);
		}
	}
	return twRC;
}


TW_UINT16  CTWAINManager::Scan(HWND hWnd, ScanCallBack pfnCallBack, void* param1, void* param2)
{
	m_pfnCallBack = NULL;
	m_pParam1 = m_pParam2 = NULL;
	TW_UINT16 twRC = TWCC_UNUSABLETWAIN;
	if(IsUsableTWAIN()){
		if((twRC =OpenDSM(hWnd)) == TWRC_SUCCESS){
			if((twRC =OpenDS()) == TWRC_SUCCESS){
				m_pfnCallBack = pfnCallBack;
				m_pParam1 = param1;
				m_pParam2 = param2;
				twRC =EnableDS(hWnd);
				m_hParentWnd = hWnd;
				if ( twRC == TWRC_SUCCESS ) {
					return twRC;
				}
				else {
					CloseDS();
					CloseDSM(hWnd);
				}
			}
			CloseDS();
		}
		CloseDSM(hWnd);
	}
	return twRC;
}


BOOL CTWAINManager::TranslateMessage(HWND hWnd, MSG* pMsg)
{
	TW_UINT16 twRC = TWRC_NOTDSEVENT;
	TW_EVENT twEvent;
	memset(&twEvent, 0, sizeof(TW_EVENT));
	if(m_bEnableDS){
		twEvent.pEvent = (TW_MEMREF)pMsg;
		twEvent.TWMessage = MSG_NULL;
		twRC = CallDSMEntry(&m_AppID, &m_DSID, DG_CONTROL, DAT_EVENT, MSG_PROCESSEVENT, (TW_MEMREF)&twEvent);
		if(twRC == TWRC_DSEVENT){
			switch(twEvent.TWMessage){
			case MSG_XFERREADY:
				TRACE(_T("MSG_XFERREADY\n"));
				twRC = NativeXfer();
				DisableDS(hWnd);
				CloseDS();
				CloseDSM(hWnd);
				switch(twRC){
				case TWRC_FAILURE:
					MessageBox(hWnd, _T("Transfer failure"), _T("TWAIN"), MB_OK | MB_ICONEXCLAMATION);
					PostMessage( m_hParentWnd, WM_TWAIN_FAILED, 0, 0 );
					m_hParentWnd = NULL;
					break;
				case TWRC_XFERDONE:
					PostMessage( m_hParentWnd, WM_TWAIN_COMPLETE, 0, 0 );
					m_hParentWnd = NULL;
					break;
				default:
				case TWRC_CANCEL:
					PostMessage( m_hParentWnd, WM_TWAIN_CANCEL, 0, 0 );
					m_hParentWnd = NULL;
					break;
				}
				break;
			case MSG_CLOSEDSREQ://
				TRACE(twEvent.TWMessage == MSG_CLOSEDSREQ ? _T("MSG_CLOSEDSREQ\n") : _T("MSG_CLOSEDSOK\n"));
				DisableDS(hWnd);
				CloseDS();
				CloseDSM(hWnd);
				PostMessage( m_hParentWnd, WM_TWAIN_CLOSEDS, 0, 0 );
				m_hParentWnd = NULL;
				break;
			default:
				break;
			}
			return TRUE;
		}
	}
	return FALSE;
}


TW_UINT16 CTWAINManager::NativeXfer()
{
	TW_UINT32 twBitmap;
	if(m_pfnCallBack == NULL)	return TWCC_PROGRAM_ERROR;
	TW_INT16 twRC = CallDSMEntry(&m_AppID, &m_DSID, DG_IMAGE, DAT_IMAGENATIVEXFER, MSG_GET, (TW_MEMREF)&twBitmap);
	switch(twRC){
	case TWRC_XFERDONE:
		(*m_pfnCallBack)(twBitmap, m_pParam1, m_pParam2);
		break;
	case TWRC_CANCEL:
		break;
	case TWRC_FAILURE:
		break;
	}
	return twRC;

}


